"""
Parse string in to FileNamePart define in namer_types.
"""
import argparse
from pathlib import PurePath
import re
import sys
from typing import List
from loguru import logger
from namer.types import FileNameParts


DEFAULT_REGEX_TOKENS = '{_site}{_sep}{_optional_date}{_ts}{_name}{_dot}{_ext}'

def name_cleaner(name: str) -> str:
    """
    Given the name parts, following a date, but preceding the file extension, attempt to glean
    extra information and discard useless information for matching with the porndb.
    """
    # truncating cruft
    for size in ['2160p', '1080p', '720p', '4k',  '3840p']:
        name = re.sub(r"[\.\- ]"+size+r"[\.\- ]{0,1}.*", "", name)
    # remove trailing ".XXX."
    name = re.sub(r"[\.\- ]{0,1}XXX[\.\- ]{0,1}.*$", "", name)
    name = re.sub(r'\.', ' ', name)
    return name


def parser_config_to_regex(tokens: str) -> str:
    """
    ``{_site}{_sep}{_optional_date}{_ts}{_name}{_dot}{_ext}``

    ``Site - YYYY.MM.DD - TS - name.mkv``

    ```
    _sep            r'[\\.\\- ]+'
    _site           r'(?P<site>[a-zA-Z0-9\\'\\.\\-\\ ]*?[a-zA-Z0-9]*?)'
    _date           r'(?P<year>[0-9]{2}(?:[0-9]{2})?)[\\.\\- ]+(?P<month>[0-9]{2})[\\.\\- ]+(?P<day>[0-9]{2})'
    _optional_date  r'(?:(?P<year>[0-9]{2}(?:[0-9]{2})?)[\\.\\- ]+(?P<month>[0-9]{2})[\\.\\- ]+(?P<day>[0-9]{2})[\\.\\- ]+)?'
    _ts             r'((?P<trans>[T|t][S|s])'+_sep+'){0,1}'
    _name           r'(?P<name>(?:.(?![0-9]{2,4}[\\.\\- ][0-9]{2}[\\.\\- ][0-9]{2}))*)'
    _dot            r'\\.'
    _ext            r'(?P<ext>[a-zA-Z0-9]{3,4})$'
    ```
    """

    _sep=r'[\.\- ]+'
    _site=r'(?P<site>.*?)'
    _date=r'(?P<year>[0-9]{2}(?:[0-9]{2})?)[\.\- ]+(?P<month>[0-9]{2})[\.\- ]+(?P<day>[0-9]{2})'
    _optional_date=r'(?:(?P<year>[0-9]{2}(?:[0-9]{2})?)[\.\- ]+(?P<month>[0-9]{2})[\.\- ]+(?P<day>[0-9]{2})[\.\- ]+)?'
    _ts=r'((?P<trans>[T|t][S|s])'+_sep+'){0,1}'
    _name=r'(?P<name>(?:.(?![0-9]{2,4}[\.\- ][0-9]{2}[\.\- ][0-9]{2}))*)'
    _dot=r'\.'
    _ext=r'(?P<ext>[a-zA-Z0-9]{3,4})$'
    regex = tokens.format_map(
        {'_site':_site,
        '_date':_date,
        '_optional_date':_optional_date,
        '_ts':_ts,
        '_name':_name,
        '_ext':_ext,
        '_sep':_sep,
        '_dot':_dot}
        )
    return regex


def parse_file_name(filename: str, regex_config: str = DEFAULT_REGEX_TOKENS) -> FileNameParts:
    """
    Given an input name of the form site-yy.mm.dd-some.name.part.1.XXX.2160p.mp4,
    parses out the relevant information in to a structure form.
    """
    regex = parser_config_to_regex(regex_config)
    file_name_parts = FileNameParts()
    file_name_parts.extension = PurePath(filename).suffix[1:]
    match = re.search(regex,filename)
    if match:
        if match.groupdict().get('year') is not None:
            prefix = "20" if len(match.group('year'))==2 else ""
            file_name_parts.date = prefix+match.group('year')+"-"+match.group('month')+"-"+match.group('day')
        if match.groupdict().get('name') is not None:
            file_name_parts.name = name_cleaner(match.group('name'))
        if match.groupdict().get('site') is not None:
            file_name_parts.site = re.sub(r'[\.\-\ ]','',match.group('site')).lower()
        if match.groupdict().get('trans') is not None:
            trans = match.group('trans')
            file_name_parts.trans = (not trans is None) and (trans.strip().upper() == 'TS')
        file_name_parts.extension = match.group('ext')
        file_name_parts.source_file_name = filename
    else:
        logger.warning("Could not parse file name: {filename}")
    return file_name_parts


def main(arglist: List[str]):
    """
    Attempt to parse a name.
    """
    description = ("You are using the file name parser of the Namer project.  " +
        "Expects a single input, and will output the contents of FileNameParts, which is the internal input " +
        "to the namer_metadatapi.py script. "+
        "Output will be the representation of that FileNameParts.\n")
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("-f", "--file", help="String to parse for name parts", required=True)
    args = parser.parse_args(arglist)
    print(parse_file_name(args.file, DEFAULT_REGEX_TOKENS))

if __name__ == "__main__":
    main(arglist=sys.argv[1:])



"""
Updates mp4 files with metadata tags readable by Plex and Apple TV App.
"""

from pathlib import Path
from loguru import logger
from mutagen.mp4 import MP4, MP4Cover, MP4StreamInfoError
from namer.types import LookedUpFileInfo, NamerConfig
from namer.ffmpeg import attempt_fix_corrupt, get_resolution, update_audio_stream_if_needed

def resolution_to_hdv_setting(resolution: int) -> int:
    """
    Using the resolution (height) of an video stream return the atom value for hdvideo
    """
    if resolution is None:
        return 0
    if resolution >= 2160:
        return 3
    if resolution >= 1080:
        return 2
    if resolution >= 720:
        return 1
    return 0


def set_if_not_none(video: MP4, atom: str, value: str):
    """
    Set a single atom on the video if it is not None.
    """
    video[atom] = [value] if value is not None else []

def get_mp4_if_possible(mp4: Path) -> MP4:
    """
    Attempt to read an mp4 file to prepare for edit.
    """
    video: MP4 = None
    try:
        video = MP4(mp4)
    except MP4StreamInfoError as __mp4_exception:
        attempt_fix_corrupt(mp4)
        video = MP4(mp4)
    return video


@logger.catch
def update_mp4_file(mp4: Path, looked_up: LookedUpFileInfo, poster: Path, config: NamerConfig):
    # pylint: disable=too-many-statements
    """
    us-tv|TV-MA|600|
    us-tv|TV-14|500|
    us-tv|TV-PG|400|
    us-tv|TV-G|300|
    us-tv|TV-Y|200|
    us-tv|TV-Y7|100|
    us-tv||0|

    mpaa|UNRATED|600|
    mpaa|NC-17|500|
    mpaa|R|400|
    mpaa|PG-13|300|
    mpaa|PG|200|
    mpaa|G|100|
    mpaa|XXX|0|
    """

    logger.info("Updating audio and tags for: {}",mp4)
    success = update_audio_stream_if_needed(mp4, config.language)
    if not success:
        logger.info("Could not process audio or copy {}",mp4)
    logger.info("Updating atom tags on: {}",mp4)
    if mp4 is not None and mp4.exists():
        video: MP4 = get_mp4_if_possible(mp4)
        set_if_not_none(video, "\xa9nam", looked_up.name)
        video["\xa9day"] = [looked_up.date+"T09:00:00Z"] if looked_up.date is not None else []
        if config.enable_metadataapi_genres:
            video["\xa9gen"] = looked_up.tags if looked_up.tags is not None else []
        else:
            video["keyw"] = looked_up.tags if looked_up.tags is not None else []
            set_if_not_none(video, "\xa9gen", config.default_genre)
        set_if_not_none(video, "tvnn", looked_up.site)
        set_if_not_none(video, "\xa9alb", looked_up.site)
        video["stik"] = [9] #Movie
        resolution = resolution_to_hdv_setting(get_resolution(mp4))
        set_if_not_none(video, "hdvd", resolution)
        set_if_not_none(video, "ldes", looked_up.description)
        video["----:com.apple.iTunes:iTunEXTC"] = 'mpaa|XXX|0|'.encode("UTF-8", errors="ignore")
        itunes_movie = '<?xml version="1.0" encoding="UTF-8"?><plist version="1.0"><dict>'
        itunes_movie += f'<key>copy-warning</key><string>{looked_up.source_url}</string>'
        itunes_movie += f'<key>studio</key> <string>{looked_up.site}</string>'
        itunes_movie += f'<key>tpdbid</key> <string>{looked_up.look_up_site_id}</string>'
        itunes_movie += '<key>cast</key> <array>'
        for performer in looked_up.performers:
            if performer.name:
                itunes_movie += f'<dict> <key>name</key> <string>{performer.name}</string>'
                if performer.role:
                    itunes_movie += f'<key>role</key> <string>{performer.role}</string>'
                itunes_movie += '</dict>'
        itunes_movie += '</array>'
        itunes_movie += '<key>codirectors</key> <array></array>'
        itunes_movie += '<key>directors</key> <array></array>'
        itunes_movie += '<key>screenwriters</key><array></array>'
        itunes_movie += '</dict></plist>'
        video["----:com.apple.iTunes:iTunMOVI"] = itunes_movie.encode("UTF-8", errors="ignore")
        if poster is not None:
            with open(poster, "rb") as file:
                ext = poster.suffix.upper()
                imageformat = None
                if ext in ['.JPEG', '.JPG']:
                    imageformat=MP4Cover.FORMAT_JPEG
                elif ext == '.PNG':
                    imageformat=MP4Cover.FORMAT_PNG
                if imageformat:
                    video["covr"] = [
                        MP4Cover(file.read(), imageformat)
                    ]
        video.save()
        logger.info("Updated atom tags: {}", mp4)
    else:
        logger.warning("Can not update tags of a non-existant file: {}", mp4)



"""
A file watching service to rename movie files and move them
to revelant locations after match the file against the porndb.
"""

import re
import shutil
import tempfile
import time
import os
import sys
from pathlib import Path, PurePath
from typing import List
from loguru import logger
from watchdog.observers.polling import PollingObserver
from watchdog.events import PatternMatchingEventHandler, FileSystemEvent, EVENT_TYPE_DELETED, EVENT_TYPE_MOVED
import schedule
from namer.namer import add_extra_artifacts, move_to_final_location, process_file
from namer.types import NamerConfig, default_config, write_log_file

def done_copying(file: Path) -> bool:
    """
    Determines if a file is being copied by checking it's size in 2 second
    increments and seeing if the size has stayed the same.
    """
    if file is None or file.exists() is False:
        return False
    while True:
        try:
            #pylint: disable=consider-using-with
            file = open(file, mode='rb')
            file.close()
            break
        except PermissionError:
            time.sleep(.2)
    return True

@logger.catch
def handle(target_file: Path, namer_config: NamerConfig):
    """
    Responsible for processing and moving new movie files.
    """
    if not target_file.exists() and target_file.is_file():
        return
    relative_path = target_file.relative_to(namer_config.watch_dir)

    # is in a dir:
    detected = PurePath(relative_path).parts[0]
    dir_path = namer_config.watch_dir / detected

    to_process = None
    workingdir = None
    workingfile = None
    if dir_path.is_dir():
        workingdir = Path(namer_config.work_dir) / detected
        logger.info("Moving {} to {} for processing", dir_path, workingdir)
        shutil.move(dir_path,workingdir)
        to_process = workingdir
    else:
        workingfile = Path(namer_config.work_dir) / relative_path
        target_file.rename(workingfile)
        logger.info("Moving {} to {} for processing", target_file, workingfile)
        to_process = workingfile
    result = process_file(to_process, namer_config)

    if result.new_metadata is None:
        if workingdir is not None:
            workingdir.rename(namer_config.failed_dir/ detected)
            logger.info("Moving failed processing {} to {} to retry later", workingdir, namer_config.failed_dir/ detected)
        else:
            newvideo = namer_config.failed_dir / relative_path
            workingfile.rename(newvideo)
            logger.info("Moving failed processing {} to {} to retry later", workingfile, newvideo)
            write_log_file(newvideo, result.search_results, namer_config)
    else:
        # Move the directory if desired.
        if (
            len(PurePath(result.final_name_relative).parts) > 1
            and workingdir is not None
            and namer_config.del_other_files is False
        ):
            target = namer_config.dest_dir / PurePath(result.final_name_relative).parts[0]
            if not target.exists():
                shutil.move(workingdir, target)
                logger.info("Moving success processed dir {} to {}", workingdir, target)
                result.video_file = namer_config.dest_dir / result.final_name_relative
        # Rename the file to dest name.
        newfile = move_to_final_location(
            result.video_file,
            namer_config.dest_dir,
            namer_config.new_relative_path_name,
            result.new_metadata,
            namer_config)
        result.video_file = newfile
        logger.info("Moving success processed file {} to {}", result.video_file, newfile)

        # Delete the workingdir if it still exists.
        if workingdir is not None and workingdir.exists():
            shutil.rmtree(workingdir, ignore_errors=True)

        # Generate/aggregate extra artifacts if desired.
        add_extra_artifacts(result, namer_config)



def retry_failed(namer_config: NamerConfig):
    """
    Moves the contents from the failed dir to the watch dir to attempt reprocessing.
    """
    logger.info("Retry failed items:")
    for file in list(namer_config.failed_dir.iterdir()):
        shutil.move( namer_config.failed_dir / file.name, namer_config.watch_dir / file.name )

def is_fs_case_sensitive():
    """
    Create a temporary file to determine if a filesystem is case sensitive, or not.
    """
    with tempfile.NamedTemporaryFile(prefix='TmP') as tmp_file:
        return(not os.path.exists(tmp_file.name.lower()))

class MovieEventHandler(PatternMatchingEventHandler):
    """
    When a new movie file is detected, this class handles the event,
    and passes off the processing to the handler() function above.
    """
    namer_config: NamerConfig

    def __init__(self, namer_config: NamerConfig):
        super().__init__(patterns=["**/*.mp4", "**/*.mkv", "**/*.MP4", "**/*.MKV"],
                         case_sensitive=is_fs_case_sensitive(), ignore_directories=True, ignore_patterns=None)
        self.namer_config = namer_config

    def on_any_event(self, event: FileSystemEvent):
        file_path = None
        if event.event_type == EVENT_TYPE_MOVED:
            file_path = event.dest_path
        elif event.event_type != EVENT_TYPE_DELETED:
            file_path = event.src_path
        if file_path is not None:
            path = Path(file_path)
            relative_path = str(path.relative_to(self.namer_config.watch_dir))
            logger.info("watchdog process called for {}", relative_path)
            if (re.search(self.namer_config.ignored_dir_regex, relative_path) is None
                and path.exists()
                and done_copying(path)
                and (path.stat().st_size / (1024*1024) > self.namer_config.min_file_size)):
                # Extra wait time in case other files are copies in as well.
                if self.namer_config.del_other_files is True:
                    time.sleep(self.namer_config.extra_sleep_time)
                handle(path, self.namer_config)

class MovieWatcher:
    """
    Watches a configured dir for new files and attempts
    to process them and move to a destination dir.

    If a failure occures moves the new files to a failed dir.

    See NamerConfig
    """
    def __init__(self, namer_config: NamerConfig):
        self.__namer_config = namer_config
        self.__src_path = namer_config.watch_dir
        self.__event_handler = MovieEventHandler(namer_config)
        self.__event_observer = PollingObserver()

    def run(self):
        """
        Checks for new files in 3 second intervals,
        needed if running in docker as events aren't properly passed in.
        """
        self.start()
        try:
            while True:
                schedule.run_pending()
                time.sleep(3)
        except KeyboardInterrupt:
            self.stop()

    def start(self):
        """
        starts a background thread to check for files.
        """
        config = self.__namer_config
        logger.info("Start porndb scene watcher.... watching: {}",config.watch_dir)
        if os.environ.get('PROJECT_VERSION'):
            project_version = os.environ.get('PROJECT_VERSION')
            print(f"Namer version: {project_version}")
        if os.environ.get('BUILD_DATE'):
            build_date = os.environ.get('BUILD_DATE')
            print(f"Built on: {build_date}")
        if os.environ.get('GIT_HASH'):
            git_hash = os.environ.get('GIT_HASH')
            print(f"Git Hash: {git_hash}")
        self.__schedule()
        self.__event_observer.start()
        # touch all existing movie files.
        files: List[Path] = []
        for file in self.__namer_config.watch_dir.rglob('*.*'):
            if file.is_file() and file.suffix in ('.mkv', '.mp4'):
                files.append(file)
        for file in files:
            if file.exists() and file.is_file():
                handle(file, self.__namer_config)

    def stop(self):
        """
        stops a background thread to check for files.
        Waits for the event processor to complete.
        """
        logger.info("exiting")
        self.__event_observer.stop()
        self.__event_observer.join()
        logger.info("exited")

    def __schedule(self):
        self.__event_observer.schedule(
            self.__event_handler,
            self.__src_path,
            recursive=True
        )


def create_watcher(namer_watchdog_config: NamerConfig) -> MovieWatcher:
    """
    Configure and start a watchdog looking for new Movies.
    """
    logger.remove()
    logger.add(sys.stdout, format="{time} {level} {message}", level="INFO")
    logger.info(str(namer_watchdog_config))
    if not namer_watchdog_config.verify_watchdog_config():
        sys.exit(-1)
    if namer_watchdog_config.retry_time is not None:
        schedule.every().day.at(namer_watchdog_config.retry_time).do(lambda: retry_failed(namer_watchdog_config))
    movie_watcher = MovieWatcher(namer_watchdog_config)
    return movie_watcher


if __name__ == "__main__":
    create_watcher(default_config()).run()
